﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Model;
using View;

namespace Controller
{
    /// <summary>
    /// Contains functions pertaining to the general game.
    /// </summary>

    /**
    * This class holds useable functions in the scope of the entire gameplay.
    */
    static class GameFunction
    {
        /**
        Returns whether an enemy unit is in range of the specified unit.
            \param graph Graph representing the current game map.
            \param unit Unit to find enemies for.
        */
        public static bool isAnEnemyUnitInRange(Graph graph, Unit unit)
        {
            if (enemyUnitsInRange(graph, unit).Count == 0)
            {
                return false;
            }
            return true;
        }

        /**
        Returns whether or not the specified enemy unit is within attack range of the specified unit.
            \param graph Graph representing the current game map.
            \param unit Specified playable unit.
            \param enemyUnit Specified enemy unit.
        */
        public static bool isEnemyUnitInRange(Graph graph, Unit unit, Unit enemyUnit)
        {
            Node attackNode = graph.getNode(enemyUnit.Position.Item1, enemyUnit.Position.Item2);
            return isNodeAttackable(unit, unit.Position.Item1, unit.Position.Item2, attackNode);
        }

        /**
        Must be called upon the start of a new turn. Resets all unit buttons and game states.
            \param player Player of the new turn.
        */
        public static void startTurn(Player player)
        {
            // set all buttons to active
            foreach (Unit unit in player.getUnits())
            {
                Button[] buttons = unit.getButtons();
                for (int i = 0; i < buttons.Count(); i++)
                {
                    if (buttons[i].getButtonType() != ButtonType.Attack
                        || buttons[i].getButtonType() != ButtonType.AttackConfirm)
                    {
                        buttons[i].Active = true;
                    }
                }
            }

            // sets all game states
            GameState.playableUnitSelected = false;
            GameState.selectedUnit = null;
            GameState.unitToAttack = null;
            GameState.dropDownMenuOpen = false;
            GameState.attackConfirmOpen = false;
            GameState.beforeMove = true;
        }

        /**
        Returns all enemy units in strict attack range of the specified unit.
            \param graph Graph representing the current game map.
            \param unit Unit to find enemy units for.
        */
        public static LinkedList<Unit> enemyUnitsInRange(Graph graph, Unit unit)
        {
            //TODO: int range = unit.getEquippedWeapon().getRange();    // use this to determine all hitable squares

            LinkedList<Unit> enemyUnitsInRange = new LinkedList<Unit>();
            LinkedList<Unit> allEnemyUnits = GameState.enemyPlayer.getUnits();

            foreach (Unit enemyUnit in allEnemyUnits)
            {
                if (isEnemyUnitInRange(graph, unit, enemyUnit))
                {
                    enemyUnitsInRange.AddLast(enemyUnit);
                }
            }

            return enemyUnitsInRange;
        }

        /**
        Returns whether or not the specified unit can perform actions.
            \param unit Specified unit.
        */
        public static bool hasUnitFinishedActions(Unit unit)
        {
            if (!unit.isButtonActive(ButtonType.AttackConfirm))
            {
                return true;
            }
            if (!unit.isButtonActive(ButtonType.Wait))
            {
                return true;
            }
            return false;
        }

        /**
        Returns whether or not the current turn is over.
        */
        public static bool isTurnOver()
        {
            foreach (Unit unit in GameState.currentPlayer.getUnits())
            {
                if (!hasUnitFinishedActions(unit))
                {
                    return false;
                }
            }
            return true;
        }

        /**
        Returns whether or not the game is over, based off win conditions.
        */
        public static bool isGameOver(Player player1, Player player2)
        {
            // if player1 does not have anymore units, game is over
            if (player1.getNumOfUnits() <= 0)
            {
                return true;
            }

            // if player2 does not have anymore units, game is over
            if (player2.getNumOfUnits() <= 0)
            {
                return true;
            }

            return false;
        }

        /**
        Checks whether or not a specified node is attackable based on the unit's equipped weapon.
            \param unit Unit to be checked for attackable nodes.
            \param unitX The x coordinate of the unit (by nodes).
            \param unitY The y coordinate of the unit (by nodes).
            \param attackNode The specified attack node.
        */
        private static bool isNodeAttackable(Unit unit, int unitX, int unitY, Node attackNode)
        {
            int attackX = attackNode.getPositionX();    // coordinates of attack node
            int attackY = attackNode.getPositionY();
            int[] range = unit.equippedWeapon.range;    // range of unit's equipped weapon

            if (attackNode.isObstacle)
            {
                return false;   // if attack node is an obstacle, automatically non-attackable
            }

            // if attack node is within weapon's min and max attack range, then node is attackable
            if (Math.Abs(unitX - attackX) + Math.Abs(unitY - attackY) >= range[0] &&
                Math.Abs(unitX - attackX) + Math.Abs(unitY - attackY) <= range[1])
            {
                return true;
            }

            return false;
        }

        /**
        Returns a list of nodes that the unit can move onto.
            \param graph Graph representing the current game map.
            \param Specified unit.
        */
        public static LinkedList<Node> setMovableNodes(Graph graph, Unit unit)
        {
            LinkedList<Node> moveableNodes = new LinkedList<Node>();
            Node currentNode = graph.getNode(unit.Position);

            // iterate through all nodes in the graph
            for (int x = 0; x < graph.Width; x++)
            {
                for (int y = 0; y < graph.Height; y++)
                {
                    // if a path exists to that node, add it to list of moveable nodes
                    if (pathFinder(graph, unit, currentNode, graph.getNode(x, y)) != null && !graph.getNode(x, y).isOccupied())
                    {
                        moveableNodes.AddLast(graph.getNode(x, y));
                    }
                }
            }
            GameState.moveableNodes = moveableNodes;
            return moveableNodes;
        }

        /**
        Returns a list of nodes that the unit can perform an attack on. Assumes the unit has not yet moved.
            \param graph Graph representing the current game map.
            \param unit Specfied unit.
        */
        public static LinkedList<Node> getAttackableNodes(Graph graph, Unit unit)
        {
            LinkedList<Node> attackableNodes = new LinkedList<Node>();

            // determine attackable nodes for each moveable node
            foreach (Node moveableNode in GameState.moveableNodes)
            {
                // iterate through entire grid to determine attackable nodes for moveable node
                for (int x = 0; x < graph.Width; x++)
                {
                    for (int y = 0; y < graph.Height; y++)
                    {
                        if ((isNodeAttackable(unit, moveableNode.getPositionX(),
                            moveableNode.getPositionY(), graph.getNode(x, y))) && (!attackableNodes.Contains(graph.getNode(x, y)))) //make sure attackable nodes is only added once each to list
                        {
                            // if node is attackable, add it to list of attackable nodes
                            attackableNodes.AddLast(graph.getNode(x, y));
                        }
                    }
                }
            }
            return attackableNodes;
        }

        /**
        Returns whether or not the specified enemy unit is within attack range of the specified unit. Assumes the unit has already moved.
            \param graph Graph representing the current game map.
            \param Specified unit.
        */
        public static LinkedList<Node> getAttackRangeAfterMoving(Graph graph, Unit unit)
        {
            LinkedList<Node> attackableNodes = new LinkedList<Node>();
            Tuple<int, int> currentPosition = unit.Position;
            for (int x = 0; x < graph.Width; x++)
            {
                for (int y = 0; y < graph.Height; y++)
                {
                    if (isNodeAttackable(unit, currentPosition.Item1, currentPosition.Item2, graph.getNode(x, y)))
                    {
                        // if node is attackable, add it to list of attackable nodes
                        attackableNodes.AddLast(graph.getNode(x, y));
                    }
                }
            }
            return attackableNodes;
        }

        /**
        Returns a list of nodes representing the path from start node to end node; if no path is valid, return null.
            \param graph Graph representing the current game map.
            \param unit Unit to move.
            \param start Start Node of the path.
            \param end End Node of the path
        */
        public static LinkedList<Node> pathFinder(Graph graph, Unit unit, Node start, Node end)
        {
            // TODO: (OPTIONAL, IGNORE FOR NOW) incorporate movability of nodes in the calculation

            if (end.isObstacle || end.isOccupied())
            {
                return null;
            }

            // uses a variation of breadth-first search
            Queue<LinkedList<Node>> paths = new Queue<LinkedList<Node>>();  // queue of paths

            // sets up queue of paths
            LinkedList<Node> startPath = new LinkedList<Node>();
            startPath.AddFirst(start);      // adds the starting node to starting path
            paths.Enqueue(startPath);

            while (paths.Count != 0)    // while queue isn't empty
            {
                LinkedList<Node> path = paths.Dequeue();    // get the first path from the queue
                Node lastNode = path.Last();                // get the last node from the path
                if (lastNode == end)                        // if last node is desired end node, return path
                {
                    return path;
                }
                if (path.Count <= unit.getMovability())     // only consider paths within the unit's movability
                {
                    // add paths with adjacent nodes of last node to queue
                    foreach (Node adjacentNode in getAdjacentNodes(graph, lastNode))
                    {
                        if (!path.Contains(adjacentNode))   // if adjacent node doesn't already exist in the path
                        {
                            LinkedList<Node> newPath = new LinkedList<Node>();
                            // copies path to new path
                            foreach (Node node in path)
                            {
                                newPath.AddLast(node);
                            }
                            // add adjacent node to new path that will be enqueued
                            newPath.AddLast(adjacentNode);
                            paths.Enqueue(newPath);
                        }
                    }
                }
            }
            return null;
        }

        /**
        Returns moveable adjacent nodes of the specified node. Used for BFS path finding.
            \param graph Graph representing the current game map.
            \param Node to get moveable adjacent nodes of.
        */
        private static LinkedList<Node> getAdjacentNodes(Graph graph, Node node)
        {
            LinkedList<Node> adjacentNodes = new LinkedList<Node>();
            Node upNode = graph.getNode(node.getPositionX(), node.getPositionY() + 1);
            Node downNode = graph.getNode(node.getPositionX(), node.getPositionY() - 1);
            Node rightNode = graph.getNode(node.getPositionX() + 1, node.getPositionY());
            Node leftNode = graph.getNode(node.getPositionX() - 1, node.getPositionY());

            // up node
            if (!upNode.isOccupied() && !upNode.isObstacle)
            {
                adjacentNodes.AddLast(upNode);
            }

            // down node
            if (!downNode.isOccupied() && !downNode.isObstacle)
            {
                adjacentNodes.AddLast(downNode);
            }

            // left node
            if (!leftNode.isOccupied() && !leftNode.isObstacle)
            {
                adjacentNodes.AddLast(leftNode);
            }

            // right node
            if (!rightNode.isOccupied() && !rightNode.isObstacle)
            {
                adjacentNodes.AddLast(rightNode);
            }

            return adjacentNodes;
        }

        /**
        Removes the specified unit from the game.
            \param unit unit to remove from the game.
        */
        public static void removeUnit(Graph graph, Player player, Unit unit)
        {
            graph.getNode(unit.Position).unitOnNode = null;
            player.removeUnit(unit);
        }
    }
}